/*
 * odeiavir.c 0.2.0 -  scans e-mail messages for viruses with the help
 *                     of an external AV program.
 *
 * Copyright (C) 2001  ViaLink Internet S/C Ltda, Rio de Janeiro, Brazil
 *                     http://www.vialink.com.br/
 * Author:             Juan Carlos Castro y Castro <jcastro@vialink.com.br>
 *
 * See the README file for more info.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, you can either send email to this
 * program's maintainer or write to: The Free Software Foundation,
 * Inc.; 59 Temple Place, Suite 330; Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <dirent.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>

#include "parseconfig.h"
#include "openantivirus.h"
#include "odeiavir.h"

/* Antivirus support - always keep AV_* constants in sync with AVDefs[] !! */

#define	AV_MCAFEE	0
#define	AV_FPROT	1
#define	AV_PANDA	2
#define AV_SOPHOS	3
#define AV_INOCIT	4
#define AV_TOTAL	5

/* Config variables */

int		cfgAllow_ns = 1;
int		cfgAllow_nr = 1;
int		cfgAllow_r = 1;
int		cfgAllow_w = 1;
int		cfgAllow_f = 1;
int		cfgAllow_av = 1;
int		cfgAllowOptions = 1;
int		cfgAV = -1;
int		cfgWarnSender = 1;
int		cfgWarnRecipient = 1;
char		*cfgWhoToWarn = NULL;
char		*cfgWhoWarns = NULL;
char		*cfgForcedRecip = NULL;
char		*cfgSubject = NULL;
int		cfgOpenAVPort = -1;
int		cfgOpenAVFirst = 1;
char		*cfgKeepViruses = NULL;
char		*cfgTempDir = NULL;

/* Config variables for the begin_av/end_av blocks */

char		*avcfgName = NULL;
char		*avcfgExecutable = NULL;
int		avcfgRetVal = -1;
char		*avcfgBeforeName = NULL;
char		*avcfgAfterName = NULL;
char		*avcfgSwitches = NULL;
int		avcfgSwitchesAfter = 0;
int		avcfgNeedsRipMIME = 1;

/* Config file definition */

PCFGOPTION AVOptions[];		/* Because of unavoidable forward reference */
void insertscanner(void);	/* Because of avoidable forward reference ;-P */

PCFGOPTION OVOptions[] =
{
	{"allow_ns", PCFG_BOOLEAN, &cfgAllow_ns},
	{"allow_nr", PCFG_BOOLEAN, &cfgAllow_nr},
	{"allow_r", PCFG_BOOLEAN, &cfgAllow_r},
	{"allow_w", PCFG_BOOLEAN, &cfgAllow_w},
	{"allow_f", PCFG_BOOLEAN, &cfgAllow_f},
	{"allow_av", PCFG_BOOLEAN, &cfgAllow_av},
	{"allow_options", PCFG_BOOLEAN, &cfgAllowOptions},
	{"antivirus", PCFG_INT, &cfgAV},
	{"warn_sender", PCFG_BOOLEAN, &cfgWarnSender},
	{"warn_recipient", PCFG_BOOLEAN, &cfgWarnRecipient},
	{"extra_warning", PCFG_STRING, &cfgWhoToWarn, 128},
	{"originator", PCFG_STRING, &cfgWhoWarns, 128},
	{"subject", PCFG_STRING, &cfgSubject, 128},
	{"openav_port", PCFG_INT, &cfgOpenAVPort},
	{"openav_first", PCFG_BOOLEAN, &cfgOpenAVFirst},
	{"keep_viruses", PCFG_STRING, &cfgKeepViruses, 256},
	{"temp_dir", PCFG_STRING, &cfgTempDir, 256},
	{"begin_av", PCFG_JUMPSTATE, AVOptions},
	{NULL}
};

PCFGOPTION AVOptions[] =
{
        {"name", PCFG_STRING, &avcfgName, 32},
        {"program", PCFG_STRING, &avcfgExecutable, 256},
	{"retval_virus", PCFG_INT, &avcfgRetVal},
	{"before_name", PCFG_STRING, &avcfgBeforeName, 128},
	{"after_name", PCFG_STRING, &avcfgAfterName, 128},
	{"switches", PCFG_STRING, &avcfgSwitches, 128},
	{"switches_after", PCFG_BOOLEAN, &avcfgSwitchesAfter},
	{"needs_ripmime", PCFG_BOOLEAN, &avcfgNeedsRipMIME},
	{"end_av", PCFG_JUMPSTATE, OVOptions, 0, insertscanner},
	{NULL}
};

/* Globals */

int	avtotal = 0;
int	firstav = 0;
int	lastav;
int	dismembered = 0;
time_t	rightnow;
AVDEF	*AVDefs = NULL;
char	subjline[] = SUBJECT_LINE;
char	buf[16384];
char	tmpdir[256];

/* Helper functions */

char *getvirusname(char *outputfile, AVDEF *avdef);
void sendwarning(char *mailto, char *warningtxt, char *virusname);
void sendwarnings(char *vname);
void removeallfiles(char *dirname);
int querydaemon(void);
void dismemberifneeded(void);

/* The Program itself in all Its infinite glory */

int main(int argc, char **argv)
{
	int		i, j, lidos, fdi, fdo, retval, retprogram = 0;
	char		*p, *target;

	/* Parse configuration file */
	parseconfig(SYSCONFDIR "/config", OVOptions);

	/* Parse command line */
	if (cfgAllowOptions)
	{
		for (i = 1; i < argc; i++)
		{
			if (!strcmp(argv[i], "-ns"))
			{
				if (!cfgAllow_ns)
					continue;
				cfgWarnSender = 0;	/* Don't warn sender */
			}
			else if (!strcmp(argv[i], "-nr"))
			{
				if (!cfgAllow_nr)
					continue;
				cfgWarnRecipient = 0;	/* Don't warn recip  */
			}
			else if (!strcmp(argv[i], "-r"))
			{
				i++;		  /* Force recip (for setups */
				if (!cfgAllow_r)  /* where address detection */
					continue; /* fails)                  */
				if (i < argc)
					cfgForcedRecip = argv[i];
			}
			else if (!strcmp(argv[i], "-w"))
			{
				i++;	/* Other address to warn (sysadmin?) */
				if (!cfgAllow_w)
					continue;
				if (i < argc)
					cfgWhoToWarn = argv[i];
			}
			else if (!strcmp(argv[i], "-f"))
			{
				i++;	/* Origin address of warnings */
				if (!cfgAllow_f)
					continue;
				if (i < argc)
					cfgWhoWarns = argv[i];
			}
			else if (!strcmp(argv[i], "-av"))
			{
				i++;	/* Antivirus to use (default = all) */
				if (!cfgAllow_av)
					continue;
				if (i < argc)
				{
					cfgAV = atoi(argv[i]);
				}
			}
			else if (!strcmp(argv[i], "-l"))
			{
				printf("\nScanner# Command\n");
				for (j = 0; j < avtotal; j++)
					printf("%8u %s (%s)\n", j,
					       AVDefs[j].executable,
					       (AVDefs[j].name) ?
					           AVDefs[j].name : "no name");
				printf("\n");
				exit(0);
			}
		}
	}
	/* If I defined an antivirus, it is the start and end */
	/* of the "for" loop below                            */
	if ((cfgAV >= 0) && (cfgAV <= avtotal))
		firstav = lastav = cfgAV;
	else
		lastav = avtotal - 1;

	/* Create work directories */
	snprintf(tmpdir, sizeof(tmpdir), "%s/odeiavir-%lu",
	                 cfgTempDir ? cfgTempDir : DEFAULT_TEMP_DIR,
                         (unsigned long) getpid());
	if (mkdir(tmpdir, 0700))
		return 0;
	if (chdir(tmpdir))
	{
		rmdir(tmpdir);
		return 0;
	}

	/* Dump message (available from stdin) into text file */
	fdo = creat("message.txt", 0600);
	if (fdo == -1)
	{
                chdir("..");
                rmdir(tmpdir);
                return 0;
        }
	while((lidos = read(0, buf, sizeof(buf))) > 0)
	{
		write(fdo, buf, lidos);
	}
	close(fdo);

	/* Run antivirus, dismembering the message with ripMIME */
	/* only if needed by a non-MIME-enabled scanner.        */

	/* Do we have to query the daemon first? */
	if ((cfgOpenAVPort != -1) && cfgOpenAVFirst)
	{
		dismemberifneeded();
		if (querydaemon())
			retprogram = 99;
	}

	/* Query conventional antiviruses */
	for (i = firstav; (i <= lastav) && (retprogram != 99); i++)
	{
		if (AVDefs[i].executable == NULL)
			continue;

		if (AVDefs[i].needs_ripmime)
		{
			dismemberifneeded();
			target = "parts/";
		}
		else
			target = "message.txt";

		if (AVDefs[i].where_options == OPTIONS_BEFORE)
			snprintf(buf, sizeof(buf),
			         "%s %s %s >output.txt 2>&1",
			         AVDefs[i].executable, AVDefs[i].options,
			         target);
		else
			snprintf(buf, sizeof(buf),
			         "%s %s %s >output.txt 2>&1",
			         AVDefs[i].executable, target,
			         AVDefs[i].options);
		retval = system(buf);

		/* If no virus, just return zero to signal message OK */
		if (WEXITSTATUS(retval) != AVDefs[i].retval_virus)
		{
			retprogram = 0;
		}
		/* Else send warning message to sender and/or recipient      */
		/* and/or someone else specified in command line, and tell   */
		/* local delivery to stop dead on its tracks by returning 99 */
		else
		{
			p = getvirusname("output.txt", &AVDefs[i]);
			sendwarnings(p);
			retprogram = 99;
			break;		/* Found a virus, no need to go on */
		}
		unlink("output.txt");
	}

	/* Do we have to query the daemon last? */
	if ((cfgOpenAVPort != -1) && !cfgOpenAVFirst && (retprogram != 99))
	{
		dismemberifneeded();
		if (querydaemon())
			retprogram = 99;
	}

	/* If we got a virus and are told to keep them, move message text */
	/* to designated area                                             */
	if ((cfgKeepViruses) && (retprogram == 99))
	{
		time(&rightnow);
		snprintf(buf, sizeof(buf), "%s/%lu-%u.txt", cfgKeepViruses,
		                rightnow, getpid());
		if ((rename("message.txt", buf) == -1) && (errno == EXDEV))
		{
			/* Tried to move across filesystems -- copy instead */
			fdi = open("message.txt", O_RDONLY);
			fdo = creat(buf, 0600);
			if ((fdi != -1) && (fdo != -1))
			{
				while((lidos = read(fdi, buf, sizeof(buf))) > 0)
					write(fdo, buf, lidos);
			}
			if (fdi != -1) close(fdi);
			if (fdo != -1) close(fdo);
		}
	}

	/* Final cleanup */
	if (dismembered)
	{
		removeallfiles("parts");
		rmdir("parts");
	}
	removeallfiles(".");
	chdir("..");
	rmdir(tmpdir);
	return retprogram;
}

/* Leech virus name from command-line scanner's display output. Takes   */
/* a pointer to an AVDEF structure in order to know what to search for. */
/*                                                                      */
/* Returns a pointer to a local static area. You have been warned.      */

char *getvirusname(char *outputfile, AVDEF *avdef)
{
	char		*pstart, *pend;
	FILE		*streami = fopen(outputfile, "r");
	static char	buf[160];

	if (!streami)
		return "";

	while (fgets(buf, sizeof(buf), streami))
	{
		/* Weed off CRs and LFs from the end */
		pend = &buf[strlen(buf) - 1];
		while (((*pend == '\r') || (*pend == '\n')) && (pend >= buf))
		{
			*pend-- = 0;
		}

		/* Search for virus string delimiters */
		if ((pstart = strstr(buf, avdef->beforename)) != NULL)
		{
			pstart += strlen(avdef->beforename);
			if (avdef->aftername[0])
			{
				pend = strstr(pstart, avdef->aftername);
				if (pend)
				{
					*pend = 0;
					fclose(streami);
					return pstart;
				}
			}
			else
			{
				/* No delimiter to the right, virus name */
				/* goes all the way to end of line       */
				fclose(streami);
				return pstart;
			}
		}
	}
	fclose(streami);
	return "";
}

/* Send a "virus found" warning message for a recipient (mailto).  */
/* warningtxt is the message template file -- it has to contain,   */
/* in addition to the message body, all necesary headers other     */
/* than Date:, To:, From:, and Subject:. The provided virusname    */
/* will be appended to the subject line.                           */
/*                                                                 */
/* Uses global variable "subjline". Trashes global buffer "buf".   */
/*                                                                 */
/* Headers from the infected message (saved to headers.txt as we   */
/* ran ripmime) are appended to the message body to ease tracing   */
/* infected machines. In case we did not run ripMIME because the   */
/* antivirus is smart, we parse the headers from message.txt the   */
/* old-fashioned way.                                              */
/*                                                                 */
/* Message is delivered thru qmail-inject. Maybe I should be       */
/* using sendmail, since qmail provides a peachy little compatible */
/* wrapper. I'm a lazy b*****d.                                    */

void sendwarning(char *mailto, char *warningtxt, char *virusname)
{
	int	lidos, fdi, fdo;
	FILE	*streami;

	fdo = creat("warning.txt", 0600);
	if (fdo != -1)
	{
		/* Write header: Subject */
		snprintf(buf, sizeof(buf), "Subject: %s %s\n",
		         (cfgSubject ? cfgSubject : subjline),
		         virusname);
		write(fdo, buf, strlen(buf));

		/* Write header: X-Mailer */
		snprintf(buf, sizeof(buf), "X-Mailer: OdeiaVir %s\n",
		         PROGVERSION);
		write(fdo, buf, strlen(buf));

		/* Write header: From */
		if (cfgWhoWarns)
		{
			if (strchr(cfgWhoWarns, '@'))
				snprintf(buf, sizeof(buf), "From: %s\n", cfgWhoWarns);
			else
				snprintf(buf, sizeof(buf), "From: %s@%s\n", cfgWhoWarns, getenv("HOST"));
		}
		else
			snprintf(buf, sizeof(buf), "From: webmaster@%s\n", getenv("HOST"));
		write(fdo, buf, strlen(buf));

		/* Write header: Date */
		time(&rightnow);
		strftime(buf, sizeof(buf), "Date: %a, %d %b %Y %H:%M:%S %z\n", localtime(&rightnow));
		write(fdo, buf, strlen(buf));

		write(fdo, "To: ", 4);
		if (mailto)
			write(fdo, mailto, strlen(mailto));
		write(fdo, "\n", 1);

		/* Write rest of headers & body from template */
		fdi = open(warningtxt, O_RDONLY);
		if (fdi != -1)
		{
			while((lidos = read(fdi, buf, sizeof(buf))) > 0)
				write(fdo, buf, lidos);
			close(fdi);
		}

		/* Write infected message's header as part of the body */
		if (dismembered)
		{
			fdi = open("headers.txt", O_RDONLY);
			if (fdi != -1)
			{
				while((lidos = read(fdi, buf, sizeof(buf))) > 0)
					write(fdo, buf, lidos);
				close(fdi);
			}
		}
		else
		{
			streami = fopen("message.txt", "r");
			if (streami)
			{
				while (fgets(buf, sizeof(buf), streami))
				{
					if ((buf[0] == '\n') || (buf[0] == '\r'))
						break;
					write(fdo, buf, strlen(buf));
				}
				fclose(streami);
			}
		}
		close(fdo);

		/* If a warning origin address is not specified, send */
		/* warnings from webmaster@(this_host). If it is but  */
		/* does not contain a "@", use (origin)@(this_host)   */
		if (cfgWhoWarns)
		{
			if (strchr(cfgWhoWarns, '@'))
				snprintf(buf, sizeof(buf), "%s -h -f%s <warning.txt >/dev/null 2>&1", EXEC_INJECT, cfgWhoWarns);
			else
				snprintf(buf, sizeof(buf), "%s -h -f%s@%s <warning.txt >/dev/null 2>&1", EXEC_INJECT, cfgWhoWarns, getenv("HOST"));
		}
		else
			snprintf(buf, sizeof(buf), "%s -h -fwebmaster@%s <warning.txt >/dev/null 2>&1", EXEC_INJECT, getenv("HOST"));
		system(buf);
	}
}

/* Send all pertinent "virus found" warning messages, depending on */
/* the relevant configuration variables.                           */

void sendwarnings(char *vname)
{
	if (cfgWarnSender)
		sendwarning(getenv("SENDER"), WARNING_MSG_SENDER, vname);
	if (cfgWarnRecipient)
	{
		if (cfgForcedRecip)
			sendwarning(cfgForcedRecip, WARNING_MSG_RECIP, vname);
		else
			sendwarning(getenv("RECIPIENT"), WARNING_MSG_RECIP,
			            vname);
	}
	if (cfgWhoToWarn)
		sendwarning(cfgWhoToWarn, WARNING_MSG_OTHER, vname);
}

/* Remove al files in a directory. Assumes the directory has nothing */
/* but regular files (good enough for us).                           */

void removeallfiles(char *dirname)
{
	DIR		*dirptr = opendir(dirname);
	struct dirent	*pfile;

	if (!dirptr)
		return;

	while ((pfile = readdir(dirptr)) != NULL)
	{
		/* Avoid . and .. */
		if (pfile->d_name[0] == '.')
		{
			if (pfile->d_name[1] == 0)
				continue;
			if (pfile->d_name[1] == '.')
				if (pfile->d_name[2] == 0)
					continue;
		}

		/* unlink() everithing else */
		snprintf(buf, sizeof(buf), "%s/%s", dirname, pfile->d_name);
		unlink(buf);
	}
	closedir(dirptr);
}

/* Queries an antivirus daemon to scan files in the "parts" directory,    */
/* sends the relevant warnings, and returns nonzero if a virus was found. */

int querydaemon(void)
{
	FILE		*daemon;
	char		*p;
	int		retval = 0;
	static char	dfound[] = "FOUND: ";

	daemon = contactdaemon(cfgOpenAVPort);
	if (daemon)
	{
		fprintf(daemon, "SCAN %s/parts\n", tmpdir);
		while (fgets(buf, sizeof(buf), daemon))
		{
			if (!strncmp(buf, dfound, sizeof(dfound) - 1))
			{
				/* Gotta strdup virus name, because */
				/* sendwarnings() trashes buf[].    */
				p = strdup(&buf[sizeof(dfound) - 1]);
				if (p)
				{
					removecrlf(p);
					sendwarnings(p);
					free(p);
				}
				retval = 1;
				break;
			}
		}
		fclose(daemon);
	}
	return retval;
}

/* Call ripMIME to dismember a message into its MIME parts. It only has */
/* to be done once per instance, hence the "if (!dismemberd)" test.     */

void dismemberifneeded(void)
{
	if (!dismembered)
	{
		/* Dismember message into MIME parts */
		if (!mkdir("parts", 0700))
		{
			system(EXEC_RIPMIME
			       " -i message.txt -d parts -e headers.txt "
			       ">/dev/null 2>&1");
		}
		dismembered = 1;
	}
}

/* Inserts a new virus scanner into the list according to avcfg* globals */

void insertscanner(void)
{
	AVDEF	*newav, *newdef;

	/* Test mandatory fields */
	if ((avcfgExecutable) && (avcfgBeforeName) && (avcfgRetVal != -1))
	{
		newav = malloc(sizeof(AVDEF));
		if (newav)
		{
			newav->name = avcfgName;
			newav->executable = avcfgExecutable;
			newav->beforename = avcfgBeforeName;
			newav->aftername = avcfgAfterName;
			newav->options = avcfgSwitches;
			newav->retval_virus = avcfgRetVal;
			newav->where_options = avcfgSwitchesAfter;
			newav->needs_ripmime = avcfgNeedsRipMIME;

			/* Insert new AVDEF into list */
			newdef = realloc(AVDefs, (avtotal + 1) * sizeof(AVDEF));
			if (newdef)
			{
				AVDefs = newdef;
				AVDefs[avtotal] = *newav;
				avtotal++;
				/* Make sure string members don't get */
				/* free()'d -- globals will be reused */
				/* when parsing next begin_av block   */
				avcfgName = NULL;
				avcfgExecutable = NULL;
				avcfgBeforeName = NULL;
				avcfgAfterName = NULL;
				avcfgSwitches = NULL;
			}
		}
	}
	/* Else syslog, when I implement logging */

	/* Cleanup */
	if (avcfgName) {free(avcfgName); avcfgName = NULL;}
	if (avcfgExecutable) {free(avcfgExecutable); avcfgExecutable = NULL;}
	if (avcfgBeforeName) {free(avcfgBeforeName); avcfgBeforeName = NULL;}
	if (avcfgAfterName) {free(avcfgAfterName); avcfgAfterName = NULL;}
	if (avcfgSwitches) {free(avcfgSwitches); avcfgSwitches = NULL;}
	avcfgRetVal = -1;
	avcfgSwitchesAfter = 0;
	avcfgNeedsRipMIME = 1;
}
